import * as THREE from "three";
import { Camera, Mesh, PerspectiveCamera, Raycaster, Scene, SphereGeometry, SRGBColorSpace, Vector2, WebGLRenderer } from "three";
import { OrbitControls } from "three/examples/jsm/controls/OrbitControls.js";
import { CSS2DRenderer } from "three/examples/jsm/renderers/CSS2DRenderer.js";
import Stats from "three/examples/jsm/libs/stats.module.js";

export type Animate = {
  fun: (arg: any) => any;
  content: any;
};
export default class Viewer {
  public id: string;
  public depth: number;
  public zAxisNumber!: number;
  public viewerDom!: HTMLElement;
  public renderer!: WebGLRenderer;
  public scene!: Scene;
  public camera!: PerspectiveCamera;
  public controls!: OrbitControls;
  public css2Renderer: CSS2DRenderer | undefined;
  public raycaster!: Raycaster;
  public mouse!: Vector2;
  public isDestroy = false; // 是否销毁
  public animateEventList = new Map();
  public statsControls!: Stats; // 声明球体几何
  public sphereGeometry!: SphereGeometry; // 声明完整球
  public sphere!: Mesh;
  public Sphere_Group!: THREE.Group;

  constructor(id: string) {
    this.id = id;
    this.depth = 1440;
    this.initViewer();
  }

  private initViewer() {
    this.initRenderer(); //创建渲染器
    this.initScene(); //创建场景
    this.initSceneBg(); //初始化背景
    this.initLight(); // 创建光源
    this.initCamera(); // 创建相机
    this.initControl(); // 创建控制器
    this.initCss2Renderer(); // 创建css2渲染器
    //
    // this.raycaster = new Raycaster(); // 创建射线
    // this.mouse = new Vector2(); // 创建鼠标位置
    //
    const animate = () => {
      if (this.isDestroy) return;
      requestAnimationFrame(animate);
      this.renderSphereRotate(); //渲染星球的自转
      this.css2Renderer?.render(this.scene, this.camera);
      this.updateDom();
      this.renderDom();

      // 全局的公共动画函数，添加函数可同步执行
      this.animateEventList.forEach((event) => {
        // console.log("animateEventList");
        if (event.fun && event.content) event.fun(event.content);
      });
    };
    animate();
  }

  private initRenderer() {
    // 获取画布dom
    this.viewerDom = document.getElementById(this.id) as HTMLElement;
    // 初始化渲染器
    this.renderer = new WebGLRenderer({
      logarithmicDepthBuffer: true, //true/false用于启用或禁用对数深度缓冲区。
      antialias: true, // true/false表示是否开启反锯齿
      alpha: true, // true/false 表示是否可以设置背景色透明
      precision: "mediump", // highp/mediump/lowp 表示着色精度选择
      premultipliedAlpha: true // true/false 表示是否可以设置像素深度（用来度量图像的分辨率）
      // preserveDrawingBuffer: false, // true/false 表示是否保存绘图缓冲
      // physicallyCorrectLights: true, // true/false 表示是否开启物理光照
    });
    this.renderer.clearDepth(); //清除深度缓冲区。在渲染之前，这通常用于重置深度缓冲区，以确保正确的深度测试

    // 开启模型对象的局部剪裁平面功能
    // 如果不设置为true，设置剪裁平面的模型不会被剪裁
    this.renderer.localClippingEnabled = true;

    this.renderer.shadowMap.enabled = true;
    this.renderer.shadowMap.type = THREE.PCFSoftShadowMap;
    this.renderer.outputColorSpace = SRGBColorSpace; // 可以看到更亮的材质，同时这也影响到环境贴图。
    this.viewerDom.appendChild(this.renderer.domElement);
  }

  private initScene() {
    this.scene = new Scene();
    // 在场景中添加雾的效果，Fog参数分别代表‘雾的颜色’、‘开始雾化的视线距离’、刚好雾化至看不见的视线距离’
    this.scene.fog = new THREE.Fog(0x000000, 0, 10000);
  }

  // 初始化背景（盒模型背景，视角在盒子里面，看到的是盒子内部）
  private initSceneBg() {
    new THREE.TextureLoader().load("/images/sky.png", (texture) => {
      const geometry = new THREE.BoxGeometry(this.viewerDom.clientWidth, this.viewerDom.clientHeight, this.depth); // 创建一个球形几何体 SphereGeometry
      const material = new THREE.MeshBasicMaterial({ map: texture, side: THREE.BackSide }); // 创建基础为网格基础材料
      const mesh = new THREE.Mesh(geometry, material);
      this.scene.add(mesh);
    });
  }

  private initLight() {
    const ambientLight = new THREE.AmbientLight(0xffffff, 1);
    // 右下角点光源
    const light_rightBottom = new THREE.PointLight(0x0655fd, 5, 0);
    light_rightBottom.position.set(0, 100, -200);
    this.scene.add(light_rightBottom);
    this.scene.add(ambientLight);
  }

  private initCamera() {
    const distance = this.viewerDom.clientWidth / 2 / Math.tan(Math.PI / 12);
    this.zAxisNumber = Math.floor(distance - this.depth / 2);
    // 渲染相机
    this.camera = new PerspectiveCamera(15, this.viewerDom.clientWidth / this.viewerDom.clientHeight, 1, 30000);
    this.camera.position.set(0, 0, this.zAxisNumber);
    this.camera.lookAt(new THREE.Vector3(0, 0, 0));
  }

  private initControl() {
    this.controls = new OrbitControls(this.camera as Camera, this.renderer?.domElement);
    this.controls.enabled = false; // enabled设置为true是可以使用鼠标控制视角
    // this.controls.enableDamping = false;
    // this.controls.screenSpacePanning = false; // 定义平移时如何平移相机的位置 控制不上下移动
    this.controls.minDistance = 2;
    this.controls.maxDistance = 2000;
    this.controls.addEventListener("change", () => {
      this.renderer.render(this.scene, this.camera);
    });
  }

  private initCss2Renderer() {
    this.css2Renderer = new CSS2DRenderer();
  }

  public updateDom() {
    this.controls.update();
    // // 更新参数
    this.camera.aspect = this.viewerDom.clientWidth / this.viewerDom.clientHeight; // 摄像机视锥体的长宽比，通常是使用画布的宽/画布的高
    this.camera.updateProjectionMatrix(); // 更新摄像机投影矩阵。在任何参数被改变以后必须被调用,来使得这些改变生效
    this.renderer.setSize(this.viewerDom.clientWidth, this.viewerDom.clientHeight);
    this.renderer.setPixelRatio(window.devicePixelRatio); // 设置设备像素比
  }

  // 渲染dom
  public renderDom() {
    this.renderer?.render(this.scene as Scene, this.camera as Camera);
  }

  private statsUpdate(statsControls: any) {
    statsControls.update();
  }

  // 初始化球体模型
  public initSphereModal() {
    //材质
    let material = new THREE.MeshPhongMaterial();
    material.map = new THREE.TextureLoader().load("/images/earth_bg.png");
    material.blendDstAlpha = 1;
    //几何体
    this.sphereGeometry = new THREE.SphereGeometry(50, 64, 32);
    //模型
    this.sphere = new THREE.Mesh(this.sphereGeometry, material);
    this.Sphere_Group = new THREE.Group();
    this.Sphere_Group.add(this.sphere);
    this.Sphere_Group.position.x = -400;
    this.Sphere_Group.position.y = 200;
    this.Sphere_Group.position.z = -200;
    this.scene.add(this.Sphere_Group);
  }

  // 渲染星球的自转
  private renderSphereRotate() {
    if (this.sphere) {
      this.Sphere_Group.rotateY(0.001);
    }
  }

  public addAnimate(id: string, animate: Animate) {
    this.animateEventList.set(id, animate);
  }

  public addStats() {
    if (!this.statsControls) this.statsControls = new Stats();
    this.statsControls.dom.style.position = "absolute";
    this.viewerDom.appendChild(this.statsControls.dom);

    // 添加到动画
    this.addAnimate("stats", { fun: this.statsUpdate, content: this.statsControls });
  }

  public destroy() {
    this.scene.traverse((child: any) => {
      if (child.material) {
        child.material.dispose();
      }
      if (child.geometry) {
        child.geometry.dispose();
      }
      child = null;
    });
    this.renderer.forceContextLoss();
    this.renderer.dispose();
    this.scene.clear();

    this.isDestroy = true;
  }
}
